#
#   Copyright (c) 2010-2014, MIT Probabilistic Computing Project
#
#   Lead Developers: Jay Baxter and Dan Lovell
#   Authors: Jay Baxter, Dan Lovell, Baxter Eaves, Vikash Mansinghka
#   Research Leads: Vikash Mansinghka, Patrick Shafto
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#

#!python
import os


class path():
    user_home_dir = os.environ['HOME']
    if 'WORKSPACE' in os.environ:
        user_home_dir = os.environ['WORKSPACE']
    # target installation for deployment
    remote_code_dir = os.path.join('/home/sgeadmin', 'tabular_predDB')
    # where we actually are right now
    this_dir = os.path.dirname(os.path.abspath(__file__))
    this_repo_dir = os.path.abspath(os.path.join(this_dir, '..'))
    install_script_dir = os.path.join(this_repo_dir, 'install_scripts')
    web_resources_dir = os.path.join(this_repo_dir, 'www')
    web_resources_data_dir = os.path.join(web_resources_dir, 'data')
    #
    install_ubuntu_script = os.path.join(install_script_dir,
                                         'install_ubuntu_packages.sh')
    install_boost_script = os.path.join(install_script_dir, 'install_boost.sh')
    virtualenv_setup_script = os.path.join(install_script_dir,
                                           'virtualenv_setup.sh')
    run_server_script = os.path.join(this_repo_dir, 'run_server.sh')
    run_webserver_script = os.path.join(this_repo_dir, 
                                        'run_simplehttpserver.sh')
    # server_script = os.path.join('jsonrpc_http', 'server_jsonrpc.py')
    try:
        os.makedirs(web_resources_dir)
        os.makedirs(web_resources_data_dir)
    except Exception, e:
        pass

class Hadoop():
    DEFAULT_CLUSTER = 'xdata_highmem'
    DEBUG = False
    #
    xdata_hadoop_jar_420 = "/usr/lib/hadoop-0.20-mapreduce/contrib/streaming/hadoop-streaming-2.0.0-mr1-cdh4.2.0.jar"
    xdata_hadoop_jar_412 = "/usr/lib/hadoop-0.20-mapreduce/contrib/streaming/hadoop-streaming-2.0.0-mr1-cdh4.1.2.jar"
    default_xdata_hadoop_jar = xdata_hadoop_jar_420 \
        if os.path.exists(xdata_hadoop_jar_420) else xdata_hadoop_jar_412
    default_xdata_compute_hdfs_uri = "hdfs://10.1.92.51:8020/"
    default_xdata_compute_jobtracker_uri = "10.1.92.53:8021"
    default_xdata_highmem_hdfs_uri = "hdfs://10.1.93.51:8020/"
    default_xdata_highmem_jobtracker_uri = "10.1.93.53:8021"
    #
    default_starcluster_hadoop_jar = "/usr/lib/hadoop-0.20/contrib/streaming/hadoop-streaming-0.20.2-cdh3u2.jar"
    default_starcluster_hdfs_uri = None
    default_starcluster_jobtracker_uri = None
    #
    default_localhost_hadoop_jar = default_xdata_hadoop_jar
    default_localhost_hdfs_uri = None
    default_localhost_jobtracker_uri = None
    #
    if DEFAULT_CLUSTER == 'starcluster':
      default_hadoop_jar = default_starcluster_hadoop_jar
      default_hdfs_uri = default_starcluster_hdfs_uri
      default_jobtracker_uri = default_starcluster_jobtracker_uri
    elif DEFAULT_CLUSTER == 'localhost':
      default_hadoop_jar = default_localhost_hadoop_jar
      default_hdfs_uri = default_localhost_hdfs_uri
      default_jobtracker_uri = default_localhost_jobtracker_uri
    else:
      default_hadoop_jar = default_xdata_hadoop_jar
      if DEFAULT_CLUSTER == 'xdata_compute':
        default_hdfs_uri = default_xdata_compute_hdfs_uri
        default_jobtracker_uri = default_xdata_compute_jobtracker_uri
      else:
        default_hdfs_uri = default_xdata_highmem_hdfs_uri
        default_jobtracker_uri = default_xdata_highmem_jobtracker_uri
    default_hadoop_binary = 'hadoop'
    default_engine_binary = '/user/bigdata/SSCI/hadoop_line_processor.jar'
    default_hdfs_dir = '/user/bigdata/SSCI/'
    default_output_path = 'myOutputDir'
    default_input_filename = 'hadoop_input'
    default_table_data_filename = 'table_data.pkl.gz'
    default_command_dict_filename = 'command_dict.pkl.gz'
    default_table_filename = os.path.join(path.web_resources_data_dir,
      'dha.csv')
    default_analyze_args_dict_filename = 'analyze_args_dict.pkl.gz'
    # 
    default_initialize_args_dict = dict(
        command='initialize',
        initialization='from_the_prior',
        )
    default_analyze_args_dict = dict(
        command='analyze',
        kernel_list=(),
        n_steps=1,
        c=(),
        r=(),
        max_time=-1,
        )

class s3():
    bucket_str = 'mitpcp-tabular-predDB'
    bucket_dir = ''
    ec2_credentials_file = os.path.expanduser('~/.boto')

class gdocs():
    auth_file = os.path.expanduser("~/mh_gdocs_auth")
    gdocs_folder_default = "MH"

class git():
    # repo_prefix = 'https://github.com/'
    # repo_prefix = 'git://github.com/'
    repo_prefix = 'git@github.com:'
    repo_suffix = 'mit-probabilistic-computing-project/tabular-predDB.git'
    repo = repo_prefix + repo_suffix
    branch = 'master'



#
#   Copyright (c) 2010-2014, MIT Probabilistic Computing Project
#
#   Lead Developers: Jay Baxter and Dan Lovell
#   Authors: Jay Baxter, Dan Lovell, Baxter Eaves, Vikash Mansinghka
#   Research Leads: Vikash Mansinghka, Patrick Shafto
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#

from bayesdb.client import Client
import StringIO  # for python 3, replace with io
import os
import sys
import time
import pickle


def assertGreaterThan(a, b, tol=0):
    """ Assert that a is at least tol greater than b.

    Examples: 
        >>> assertGreaterThan(7, 5, tol=0)  # passes
        >>> assertGreaterThan(7, 5, tol=1)  # passes
        >>> assertGreaterThan(7, 5, tol=5)  # fails
    """
    try:
        assert(a >= b*(1.0+tol))
    except AssertionError:
        raise AssertionError("a(%s) is not %f greater than b(%s)." % (str(a), tol*100, str(b)))


def assertLessThan(a, b, tol=0):
    """ Assert that a is at least tol less than b.

    Examples:
        >>> assertLessThan(5, 7, tol=0)  # passes
        >>> assertLessThan(5, 7, tol=1)  # passes
        >>> assertLessThan(5, 7, tol=5)  # fails
    """
    try:
        assert(a < b*(1.0-tol))
    except AssertionError:
        raise AssertionError("a(%s) is not %f less than than b(%s)." % (str(a), tol*100, str(b)))


def assertClose(a, b, tol=10E-6):
    """ Asset that a is within tol of b.

    Examples:
        >>> assertClose(2.0, 2.1, tol=.2)  # passes
        >>> assertClose(2.1, 2.0, tol=.2)  # passes
        >>> assertClose(2.0, 2.1)  # fails
    """
    try:
        assert(abs(a-b) < tol)
    except AssertionError:
        raise AssertionError("Difference (%f) between a(%f) and b(%f) if greate than %f."
                             % (abs(a-b), a, b, tol))


def assertIn(a, b):
    assert(a in b)


def assertNotIn(a, b):
    assert(a not in b)


class TestClient(Client):
    """ 
    Wrapper for Client. Test fixture for appyling unit tests to use cases and generating 
    sphinx-ready docs.

    Attributes:
        test_id (str): Name of test. Used to name files. Replaces instances pf 'btable' in sphinx 
            output.
        csv_filename (string): Name of csv datafile for btable creation
        commands (list of str): List of commands executed during btable creation as well as those 
            executed by the user.
        comments (list of str): List of comments (in .rst format) added by the user.
        comment_indices (list of int): Stores the command index of the comment. That is, 
            comment_indices[i] = j means that comments[i] should be placed before immediately 
            before command j.
        output (list of str): Stores resulting pretty output of each command.
        timestamp (str): ID used for file naming.
        btable_name (str): Name of btable used for test.
        dir (str): Name of the directory to which files are saved.

    Notes:
        See tests/example_tests/test_dha_example.py for a detailed use example.

    """
    def __init__(self, test_id, csv_filename, model_filename=None, num_models=10,
                 num_analyze_iterations=250, num_analyze_minutes=None, schema_update_commands=[],
                 key_column=0):
        """
        Create TestClient object, CREATE, INITIALIZE, ANALYZE, and (optionally) UPDATE btable, and
        fill in initial commands and output.

        Args:
            test_id (str): Name of test. Used to name files. Replaces instances pf 'btable' in sphinx 
                output.
            csv_filename (string): Name of csv datafile for btable creation

        Kwargs:
            model_filename (str): Filename of models to load. If None (default) creates and 
                analyzes new models.
            num_models (int): Number of models to create for btable.
            num_analyze_iterations (int): Number of iterations to analyze btable.
            num_analyze_minutes (int): Number of minutes to analyze btable. If not None (default), 
                overrides num_analyze_iterations.
            schema_update_commands (list of str): List of commands to run after initialization. Use
                these commands to update schema or add codebooks on start up.
            key_column (int): Key column of btable.

        Example:
            Create Tester object, analyze 100 models and update schema.
            >>> cmds = ['UPDATE SCHEMA FOR <btable> SET nbrhd=categorical']
            >>> tclient = TestClient('homes','data/homes.csv', num_models=100, schema_update_commands=cmds)
        """

        super(TestClient, self).__init__()
        self.test_id = test_id
        self.csv_filename = csv_filename
        self.commands = []
        self.comments = []
        self.comment_indices = []
        self.output = []

        # create timestamp and temp directory
        self.timestamp = '_' + str(int(time.time()*100))
        self.btable_name = test_id + self.timestamp

        self.dir = os.path.join('data', test_id, test_id+self.timestamp)
        # create datat directory
        if not os.path.exists(self.dir):
            os.makedirs(self.dir)

        self._initialize(model_filename, num_models, num_analyze_iterations, num_analyze_minutes,
                         schema_update_commands, key_column)
        # create btable

    def _initialize(self, model_filename=None, num_models=10, num_analyze_iterations=250,
                    num_analyze_minutes=None, schema_update_commands=[], key_column=0):
        commands = []
        create_btable_cmd = 'CREATE BTABLE <btable> FROM %s;' % (self.csv_filename)
        commands.append(create_btable_cmd)
        if model_filename is not None:
            load_model_cmd = 'LOAD MODELS %s INTO <btable>;' % (model_filename)
            commands.append(load_model_cmd)
        else:
            init_models_cmd = 'INITIALIZE %i MODELS FOR <btable>;' % (num_models)
            commands.append(init_models_cmd)
            commands.extend(schema_update_commands)
            if num_analyze_minutes is not None:
                analyze_cmd = 'ANALYZE <btable> FOR %i MINUTES WAIT;' % (num_analyze_minutes)
            else:
                analyze_cmd = 'ANALYZE <btable> FOR %i ITERATIONS WAIT;' % (num_analyze_iterations)
            commands.append(analyze_cmd)

        self.commands = commands

        for cmd in commands:
            # redirect print output
            out = StringIO.StringIO()
            oldstdout = sys.stdout
            sys.stdout = out

            super(TestClient, self).execute(cmd.replace('<btable>', self.btable_name), pretty=True,
                                            key_column=key_column, debug=False, yes=True,
                                            force_output=True)

            # restore normal output output
            sys.stdout = oldstdout
            self.output.append(out.getvalue())

    def finalize(self):
        """ Saves the models, drops the btable and (TODO) generates .rst output
        """
        output_model_filename = os.path.join(self.dir, self.btable_name + '_models.pkl.gz')
        super(TestClient, self).execute('SAVE MODELS FROM %s TO %s;'
                                        % (self.btable_name, output_model_filename), yes=True)
        super(TestClient, self).execute('DROP BTABLE %s;' % (self.btable_name), yes=True)
        output = {
            'test_id': self.test_id,
            'commands': self.commands,
            'comments': self.comments,
            'comment_indices': self.comment_indices,
            'output': self.output,
            'output_model_filename': output_model_filename,
            'timestamp': self.timestamp
        }

        # save sphinx output if needed
        self.build_rst()
        pickle.dump(output, open(os.path.join(self.dir, 'output.pkl'), 'wb'))

    def build_rst(self):
        """ Build .rst file and save to dir
        """
        txt_idx = 0
        cmd_idx = 0
        rst_output = ""

        while txt_idx < len(self.comments) and cmd_idx < len(self.commands):
            while self.comment_indices[txt_idx] <= cmd_idx:
                rst_output += self.comments[txt_idx].rstrip()
                txt_idx += 1
                if self.comment_indices[txt_idx] <= cmd_idx:
                    rst_output += '\n'

            rst_output += "::\n\n"

            next_cmt_idx = self.comment_indices[txt_idx]
            while cmd_idx < next_cmt_idx:
                cmd_txt = '\t>>> ' + self.commands[cmd_idx] + '\n' + '\t'
                out_text = self.output[cmd_idx].replace('\n', '\n\t').replace(self.btable_name, self.test_id) + '\n'
                rst_output += cmd_txt.replace(self.dir+'/', '') + out_text
                cmd_idx += 1

            rst_output += "\n"

        if txt_idx < len(self.comments):
            for comment in self.comments[txt_idx:]:
                rst_output += comment + '\n'

        rst_output = rst_output.replace('<btable>', self.test_id)

        with open(os.path.join(self.dir, 'output.rst'), 'wb') as f:
            f.write(rst_output)

    def comment(self, comment_text, introduction=False):
        """ Add a comment after current command

        Args:
            comment_text (str): .rst-formatted text to add. Comments are added in the order in 
                which they are supplied.

        Kwargs:
            introduction (bool): If true, adds the comment before the btable initialization 
                commands. Note that the introduction comment must be added first.
        """
        if introduction:
            if len(self.comments) > 0:
                raise ValueError("Introduction must be first comment added.")
            self.comment_indices.append(0)
        else:
            self.comment_indices.append(len(self.commands))

        self.comments.append(comment_text)

    def __call__(self, call_input, pretty=True, timing=False, wait=False, plots=None, yes=False,
                 debug=False, pandas_df=True, pandas_output=True, key_column=None,
                 return_raw_result=False, ignore_output=False, force_output=True):
        """ Calls BayesDB Client, redirects and saves output, and returns data.
        """
        # redirect print output
        out = StringIO.StringIO()
        oldstdout = sys.stdout
        sys.stdout = out

        res = super(TestClient, self).execute(call_input.replace('<btable>', self.btable_name),
                                              pretty, timing, wait, plots, yes, debug, pandas_df,
                                              pandas_output, key_column, return_raw_result,
                                              force_output)
        # restore normal output output
        sys.stdout = oldstdout
        self.output.append(out.getvalue())

        if not ignore_output:
            self.commands.append(call_input)

        if res is not None:
            return res[0]
        else:
            return None



#!python
#
#   Copyright (c) 2010-2014, MIT Probabilistic Computing Project
#
#   Lead Developers: Dan Lovell and Jay Baxter
#   Authors: Dan Lovell, Baxter Eaves, Jay Baxter, Vikash Mansinghka
#   Research Leads: Vikash Mansinghka, Patrick Shafto
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#
import os
#
from starcluster.clustersetup import ClusterSetup
from starcluster.logger import log


project_name = 'bayesdb'
#
repo_url = 'https://github.com/mit-probabilistic-computing-project/%s.git' % project_name
get_repo_dir = lambda user: os.path.join('/home', user, project_name)
get_setup_script = lambda user: os.path.join(get_repo_dir(user), 'setup.py')


class bayesdbSetup(ClusterSetup):

    def __init__(self):
        # TODO: Could be generalized to "install a python package plugin"
        pass

    def run(self, nodes, master, user, user_shell, volumes):
        # set up some paths
        repo_dir = get_repo_dir(user)
        setup_script = get_setup_script(user)
        for node in nodes:
            # NOTE: nodes includes master
            log.info("Installing %s as root on %s" % (project_name, node.alias))
            #
            cmd_strs = [
                # FIXME: do this somewhere else
                'pip install pyparsing==2.0.1',
                'pip install patsy',
                'pip install statsmodels',
                'rm -rf %s' % repo_dir,
                'git clone %s %s' % (repo_url, repo_dir),
                'python %s develop' % setup_script,
                # 'python %s build_ext --inplace' % setup_script,
                'chown -R %s %s' % (user, repo_dir),
            ]
            for cmd_str in cmd_strs:
                node.ssh.execute(cmd_str + ' >out 2>err')
                pass
            pass
        for node in nodes:
            log.info("Setting up %s as %s on %s" % (project_name, user, node.alias))
            #
            cmd_strs = [
                'mkdir -p ~/.matplotlib',
                'echo backend: Agg > ~/.matplotlib/matplotlibrc',
            ]
            for cmd_str in cmd_strs:
                node.shell(user=user, command=cmd_str)
                pass
            pass
        return



#
#   Copyright (c) 2010-2014, MIT Probabilistic Computing Project
#
#   Lead Developers: Jay Baxter and Dan Lovell
#   Authors: Jay Baxter, Dan Lovell, Baxter Eaves, Vikash Mansinghka
#   Research Leads: Vikash Mansinghka, Patrick Shafto
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#

import numpy as np
import pylab as p
import os
from matplotlib.colors import LogNorm, Normalize
from matplotlib.ticker import MaxNLocator
import matplotlib.gridspec as gs
import matplotlib.cm
import pandas
import numpy
import utils
import functions
import data_utils as du
import math

def turn_off_labels(subplot):
    subplot.axes.get_xaxis().set_visible(False)
    subplot.axes.get_yaxis().set_visible(False)


def plot_general_histogram(colnames, data, M_c, schema_full, filename=None, scatter=False,
                           remove_key=False):
    '''
    colnames: list of column names
    data: list of tuples (first list is a list of rows, so each inner tuples is a row)
    colnames = ['name', 'age'], data = [('bob',37), ('joe', 39),...]
    scatter: False if histogram, True if scatterplot
    '''
    numcols = len(colnames)
    if remove_key:
        numcols -= 1

    if numcols > 1:
        gsp = gs.GridSpec(1, 1)
        plots = create_pairwise_plot(colnames, data, M_c, schema_full, gsp, remove_key=remove_key)
    else:
        f, ax = p.subplots()
        parsed_data = parse_data_for_hist(colnames, data, M_c, schema_full, remove_key=remove_key)
        create_plot(parsed_data, ax, horizontal=True)

    if filename:
        p.savefig(filename)
        p.close()
    else:
        p.show()


def create_plot(parsed_data, subplot, label_x=True, label_y=True, text=None, compress=False,
                super_compress=False, **kwargs):
    """
    Takes parsed data and a subplot object, and creates a plot of the data on that subplot object.
    """
    if parsed_data['datatype'] == 'mult1D':
        if len(parsed_data['data']) == 0:
            return
        subplot.tick_params(top='off', bottom='off', left='off', right='off')
        subplot.axes.get_yaxis().set_ticks([])
        labels = parsed_data['labels']
        datapoints = parsed_data['data']
        num_vals = len(labels)
        ind = np.arange(num_vals)
        width = 0.5

        horizontal = 'horizontal' in kwargs and kwargs['horizontal']
        rot = (horizontal != super_compress)

        if horizontal:
            plot_method = subplot.barh
            label_method = subplot.set_ylabel
            tick_axis = subplot.axes.get_yaxis()
            hide_axis = subplot.axes.get_xaxis()
        else:
            plot_method = subplot.bar
            label_method = subplot.set_xlabel
            tick_axis = subplot.axes.get_xaxis()
            hide_axis = subplot.axes.get_yaxis()

        plot_method(ind, datapoints, width, color=matplotlib.cm.Blues(0.5), align='center')
        label_method(parsed_data['axis_label'])

        n_labels = len(labels)

        if (not compress and len(labels) < 15) or (compress and len(labels) < 5):
            tick_axis.set_ticks(range(n_labels))
            tick_axis.set_ticklabels(labels)
        if compress:
            hide_axis.set_visible(False)

    elif parsed_data['datatype'] == 'cont1D':
        if len(parsed_data['data']) == 0:
            return
        datapoints = parsed_data['data']
        subplot.series = pandas.Series(datapoints)
        horizontal = 'horizontal' in kwargs and kwargs['horizontal']

        if horizontal:
            subplot.series.hist(normed=True, color=matplotlib.cm.Blues(0.5),
                                orientation='horizontal')
            label_method = subplot.set_ylabel
            hide_axis = subplot.axes.get_xaxis()
            major_axis = subplot.axes.get_yaxis()
        else:
            subplot.series.hist(normed=True, color=matplotlib.cm.Blues(0.5))
            label_method = subplot.set_xlabel
            hide_axis = subplot.axes.get_yaxis()
            major_axis = subplot.axes.get_xaxis()
            if not compress:
                subplot.series.dropna().plot(kind='kde', style='r--')

        label_method(parsed_data['axis_label'])

        if compress:
            hide_axis.set_visible(False)
            major_axis.set_major_locator(MaxNLocator(nbins=3))

    elif parsed_data['datatype'] == 'contcont':
        if len(parsed_data['data_y']) == 0 or len(parsed_data['data_x']) == 0:
            return
        subplot.hist2d(parsed_data['data_y'], parsed_data['data_x'],
                       bins=max(len(parsed_data['data_x'])/200, 40),
                       norm=LogNorm(), cmap=matplotlib.cm.Blues)
        if not compress:
            subplot.set_xlabel(parsed_data['axis_label_x'])
            subplot.set_ylabel(parsed_data['axis_label_y'])
        else:
            turn_off_labels(subplot)

    elif parsed_data['datatype'] == 'multmult':
        if len(parsed_data['data']) == 0:
            return
        subplot.tick_params(labelcolor='b', top='off', bottom='off', left='off', right='off')
        # subplot.axes.get_xaxis().set_ticks([])
        unique_xs = parsed_data['labels_x']
        unique_ys = parsed_data['labels_y']
        dat = parsed_data['data']
        norm_a = Normalize(vmin=dat.min(), vmax=dat.max())
        subplot.imshow(parsed_data['data'], norm=Normalize(), interpolation='nearest',
                       cmap=matplotlib.cm.Blues, aspect=float(len(unique_xs))/len(unique_ys))

        subplot.axes.get_xaxis().set_ticks(range(len(unique_xs)))
        subplot.axes.get_xaxis().set_ticklabels(unique_xs, rotation=90)
        subplot.axes.get_yaxis().set_ticks(range(len(unique_ys)))
        subplot.axes.get_yaxis().set_ticklabels(unique_ys)
        if not compress:
            subplot.set_xlabel(parsed_data['axis_label_x'])
            subplot.set_ylabel(parsed_data['axis_label_y'])
        else:
            turn_off_labels(subplot)

    elif parsed_data['datatype'] == 'multcont':
        # categorical is always first. parsed_data['transpose'] is true if categorical should be on
        # Y axis.
        values = parsed_data['values']
        groups = parsed_data['groups']
        vert = not parsed_data['transpose']
        subplot.boxplot(values, vert=vert)

        if compress:
            turn_off_labels(subplot)
        else:
            if vert:
                xtickNames = p.setp(subplot, xticklabels=groups)
                p.setp(xtickNames, fontsize=8, rotation=90)
            else:
                p.setp(subplot, yticklabels=groups)

    else:
        raise Exception('Unexpected data type, or too many arguments')

    x0, x1 = subplot.get_xlim()
    y0, y1 = subplot.get_ylim()
    aspect = (abs(float((x1 - x0))) / abs(float((y1 - y0))))
    subplot.set_aspect(aspect)
    return subplot


def any_nan(row):
    return any([isinstance(x, float) and math.isnan(x) for x in row])


def parse_data_for_hist(colnames, data, M_c, schema_full, remove_key=False):
    columns = colnames[:]
    # Remove key column if present
    if remove_key:
        columns.pop(0)
        data = [row[1:] for row in data]
    # Remove any rows with nan values.
    data = [row for row in data if not any_nan(row)]
    # Stop if there are no rows remaining after cleaning missing values.
    if len(data) == 0:
        raise utils.BayesDBError('There are no datapoints that contain values from every category '
                                 'specified. Try excluding columns with many NaN values.')

    # Pull items from M_c to simplify code throughout the rest of this function
    name_to_idx = M_c['name_to_idx']
    column_metadata = M_c['column_metadata']
    cctypes = [schema_full[column] for column in columns]

    # Treat cyclic as numerical until we establish what we want in a cyclic plot.
    for cctype_idx, cctype in enumerate(cctypes):
        if cctype == 'cyclic':
            cctypes[cctype_idx] = 'numerical'

    output = {}
    if len(columns) == 1:
        np_data = np.array([x[0] for x in data])

        # Allow col_idx to be None, to allow for predictive functions to be plotted.
        if columns[0] in name_to_idx:
            col_idx = name_to_idx[columns[0]]
        else:
            col_idx = None

        # Treat not-column (e.g. function) the same as numerical, since no code to value conversion.
        if col_idx is None or cctypes[0] == 'numerical':
            output['datatype'] = 'cont1D'
            output['data'] = np_data
        elif cctypes[0] == 'categorical':
            unique_labels = sorted(column_metadata[name_to_idx[columns[0]]]['code_to_value'].keys())
            counts = []
            for label in unique_labels:
                counts.append(sum(np_data == str(label)))
            output['datatype'] = 'mult1D'
            output['labels'] = unique_labels
            output['data'] = counts

        try:
            # try to get short names from M_c_full
            short_name = M_c['column_codebook'][col_idx]['short_name']
            output['axis_label'] = short_name
            output['title'] = short_name
        except KeyError:
            output['axis_label'] = columns[0]
            output['title'] = columns[0]

    elif len(columns) == 2:
        # Treat not-column (e.g. function) the same as numerical, since no code to value conversion.
        if columns[0] in name_to_idx:
            col_idx_1 = name_to_idx[columns[0]]
        else:
            col_idx_1 = None
        if columns[1] in name_to_idx:
            col_idx_2 = name_to_idx[columns[1]]
        else:
            col_idx_2 = None

        if cctypes[0] == 'numerical' and cctypes[1] == 'numerical':
            output['datatype'] = 'contcont'
            output['data_x'] = [x[0] for x in data]
            output['data_y'] = [x[1] for x in data]

        elif cctypes[0] == 'categorical' and cctypes[1] == 'categorical':
            counts = {}  # keys are (var 1 value, var 2 value)
            # data contains a tuple for each datapoint: (value of var 1, value of var 2)
            for row in data:
                row = tuple(row)
                if row in counts:
                    counts[row] += 1
                else:
                    counts[row] = 1

            # these are the values.
            unique_xs = sorted(column_metadata[col_idx_2]['code_to_value'].keys())
            unique_ys = sorted(column_metadata[col_idx_1]['code_to_value'].keys())
            unique_ys.reverse()  # Hack to reverse the y's
            x_ordered_codes = [du.convert_value_to_code(M_c, col_idx_2, xval) for xval in unique_xs]
            y_ordered_codes = [du.convert_value_to_code(M_c, col_idx_1, yval) for yval in unique_ys]

            # Make count array: indexed by y index, x index
            counts_array = numpy.zeros(shape=(len(unique_ys), len(unique_xs)))
            for i in counts:
                # this converts from value to code
                y_index = y_ordered_codes.index(column_metadata[col_idx_1]['code_to_value'][i[0]])
                x_index = x_ordered_codes.index(column_metadata[col_idx_2]['code_to_value'][i[1]])
                counts_array[y_index][x_index] = float(counts[i])
            output['datatype'] = 'multmult'
            output['data'] = counts_array
            output['labels_x'] = unique_xs
            output['labels_y'] = unique_ys

        elif 'numerical' in cctypes and 'categorical' in cctypes:
            output['datatype'] = 'multcont'
            categories = {}

            categorical_column = cctypes.index('categorical')

            groups = sorted(column_metadata[name_to_idx[columns[categorical_column]]]['code_to_value'].keys())
            for i in groups:
                categories[i] = []
            for i in data:
                categories[i[categorical_column]].append(i[1 - categorical_column])

            output['groups'] = groups
            output['values'] = [categories[x] for x in groups]
            output['transpose'] = (categorical_column == 0)

        try:
            # try to get short names from M_c_full
            columns[0] = M_c['column_codebook'][col_idx_1]['short_name']
            columns[1] = M_c['column_codebook'][col_idx_2]['short_name']
        except KeyError:
            pass

        output['axis_label_x'] = columns[1]
        output['axis_label_y'] = columns[0]

        output['title'] = columns[0] + ' -versus- ' + columns[1]

    else:
        output['datatype'] = None

    return output


def create_pairwise_plot(colnames, data, M_c, schema_full, gsp, remove_key=False):
    columns = colnames[:]
    # Remove key column if present
    if remove_key:
        columns.pop(0)
        data = [row[1:] for row in data]
    # Remove any rows with nan values.
    data = [row for row in data if not any_nan(row)]
    # Stop if there are no rows remaining after cleaning missing values.
    if len(data) == 0:
        raise utils.BayesDBError('There are no datapoints that contain values from every category '
                                 'specified. Try excluding columns with many NaN values.')

    output = {}

    n_columns = len(columns)
    # Rotate outer labels if there are more than 6 columns to be plotted.
    super_compress = n_columns > 6
    gsp = gs.GridSpec(n_columns, n_columns)
    for i in range(n_columns):
        for j in range(n_columns):
            if j == 0 and i < n_columns - 1:
                # left hand marginals
                sub_colnames = [columns[i]]
                sub_data = [[x[i]] for x in data]
                parsed_data = parse_data_for_hist(sub_colnames, sub_data, M_c, schema_full)
                create_plot(parsed_data, p.subplot(gsp[i, j], adjustable='box', aspect=1), False,
                            False, columns[i], horizontal=True, compress=True,
                            super_compress=super_compress)

            elif i == n_columns - 1 and j > 0:
                # bottom marginals
                subdata = None
                if j == 1:
                    sub_colnames = [columns[n_columns - 1]]
                    sub_data = [[x[n_columns - 1]] for x in data]
                else:
                    sub_colnames = [columns[j - 2]]
                    sub_data = [[x[j - 2]] for x in data]
                parsed_data = parse_data_for_hist(sub_colnames, sub_data, M_c, schema_full)
                create_plot(parsed_data, p.subplot(gsp[i, j], adjustable='box', aspect=1), False,
                            False, columns[j-2], horizontal=False, compress=True,
                            super_compress=super_compress)

            elif (j != 0 and i != n_columns - 1) and j < i + 2:
                # pairwise joints
                j_col = j - 2
                if j == 1:
                    j_col = n_columns - 1
                sub_colnames = [columns[i], columns[j_col]]
                sub_data = [[x[i], x[j_col]] for x in data]
                parsed_data = parse_data_for_hist(sub_colnames, sub_data, M_c, schema_full)
                create_plot(parsed_data, p.subplot(gsp[i, j]), False, False, horizontal=True,
                            compress=True, super_compress=super_compress)
            else:
                pass


def plot_matrix(matrix, column_names, title='', filename=None):
    # actually create figure
    fig = p.figure()
    fig.set_size_inches(16, 12)
    p.imshow(matrix, interpolation='none', cmap=matplotlib.cm.Blues, vmin=0, vmax=1)
    p.colorbar()
    p.gca().set_yticks(range(len(column_names)))
    p.gca().set_yticklabels(column_names, size='small')
    p.gca().set_xticks(range(len(column_names)))
    p.gca().set_xticklabels(column_names, rotation=90, size='small')
    p.title(title)
    if filename:
        p.savefig(filename)
    else:
        fig.show()


def _create_histogram(M_c, data, columns, mc_col_indices, filename):
    dir = S.path.web_resources_data_dir
    full_filename = os.path.join(dir, filename)
    num_rows = data.shape[0]
    num_cols = data.shape[1]

    p.figure()
    # col_i goes from 0 to number of predicted columns
    # mc_col_idx is the original column's index in M_c
    for col_i in range(num_cols):
        mc_col_idx = mc_col_indices[col_i]
        data_i = data[:, col_i]
        ax = p.subplot(1, num_cols, col_i, title=columns[col_i])
        if M_c['column_metadata'][mc_col_idx]['modeltype'] == 'normal_inverse_gamma':
            p.hist(data_i, orientation='horizontal')
        else:
            str_data = [du.convert_code_to_value(M_c, mc_col_idx, code) for code in data_i]
            unique_labels = list(set(str_data))
            np_str_data = np.array(str_data)
            counts = []
            for label in unique_labels:
                counts.append(sum(np_str_data == label))
            num_vals = len(M_c['column_metadata'][mc_col_idx]['code_to_value'])
            rects = p.barh(range(num_vals), counts)
            heights = np.array([rect.get_height() for rect in rects])
            ax.set_yticks(np.arange(num_vals) + heights/2)
            ax.set_yticklabels(unique_labels)

    p.tight_layout()
    p.savefig(full_filename)



#
#   Copyright (c) 2010-2014, MIT Probabilistic Computing Project
#
#   Lead Developers: Jay Baxter and Dan Lovell
#   Authors: Jay Baxter, Dan Lovell, Baxter Eaves, Vikash Mansinghka
#   Research Leads: Vikash Mansinghka, Patrick Shafto
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#

import time
import inspect
import pickle
import os
import numpy
import pytest
import random
import shutil
import pandas

import bayesdb.utils as utils
from bayesdb.client import Client
from bayesdb.engine import Engine
import bayesdb.bql_grammar as bql
'''
Table data for: 
+--------+-----+-----+----+-----+
| row_id |  a  |  b  | c  |  d  |
+--------+-----+-----+----+-----+
|   0    |  1  | 1.0 | we | nan |
|   1    |  2  | 2.0 | we |  2  |
|   2    |  a  | 3.0 | we |  4  |
|   3    |  4  | 3.0 | w  | nan |
|   4    |  4  | 4.0 | w  |  5  |
|   5    |  6  | 5.0 | e  | nan |
|   6    | nan | 6.0 | w  |  3  |
|   7    | nan | 7.0 | sd |  3  |
|   8    | nan | 7.0 | sd |  3  |
+--------+-----+-----+----+-----+

created from csv: 
a,b,c,d
1,1,we,
2,2,we,2
a,3,we,4
4,3,w,
4,4,w,5
6,5,e,
,6,w,3
,7,sd,3
,7,sd,3

'''
test_M_c = {'idx_to_name': {'1': 'b', '0': 'a', '3': 'd', '2': 'c'},
            'column_metadata': [
                    {'code_to_value': {'a': 0, '1': 1, '2': 2, '4': 3, '6': 4}, 
                     'value_to_code': {0: 'a', 1: '1', 2: '2', 3: '4', 4: '6'}, 
                     'modeltype': 'symmetric_dirichlet_discrete'}, 
                    {'code_to_value': {}, 'value_to_code': {}, 
                     'modeltype': 'normal_inverse_gamma'}, 
                    {'code_to_value': {'we': 0, 'e': 1, 'w': 2, 'sd': 3}, 
                     'value_to_code': {0: 'we', 1: 'e', 2: 'w', 3: 'sd'}, 
                     'modeltype': 'symmetric_dirichlet_discrete'}, 
                    {'code_to_value': {'3': 1, '2': 2, '5': 0, '4': 3}, 
                     'value_to_code': {0: '5', 1: '3', 2: '2', 3: '4'}, 
                     'modeltype': 'symmetric_dirichlet_discrete'}], 
            'name_to_idx': {'a': 0, 'c': 2, 'b': 1, 'd': 3}}

test_T = [[1.0, 1.0, 0.0, numpy.nan], 
          [2.0, 2.0, 0.0, 2.0], 
          [0.0, 3.0, 0.0, 3.0], 
          [3.0, 3.0, 2.0, numpy.nan], 
          [3.0, 4.0, 2.0, 0.0], 
          [4.0, 5.0, 1.0, numpy.nan], 
          [numpy.nan, 6.0, 2.0, 1.0], 
          [numpy.nan, 7.0, 3.0, 1.0], 
          [numpy.nan, 7.0, 3.0, 1.0]]

def test_row_id_from_col_value():
    assert utils.row_id_from_col_value('1', 'a', test_M_c, test_T) == 0
    assert utils.row_id_from_col_value('7', 'a', test_M_c, test_T) == None
    assert utils.row_id_from_col_value('1', 'b', test_M_c, test_T) == 0
    assert utils.row_id_from_col_value(1, 'b', test_M_c, test_T) == 0
    assert utils.row_id_from_col_value('1', 'b', test_M_c, test_T) == 0
    try:
        utils.row_id_from_col_value('4', 'a', test_M_c, test_T)
        assert False
    except utils.BayesDBUniqueValueError:
        assert True

def test_string_to_column_type():
    assert utils.string_to_column_type('1', 'a', test_M_c) == '1'
    assert utils.string_to_column_type('1', 'b', test_M_c) == 1
    assert utils.string_to_column_type(1, 'a', test_M_c) == 1



#
#   Copyright (c) 2010-2014, MIT Probabilistic Computing Project
#
#   Lead Developers: Jay Baxter and Dan Lovell
#   Authors: Jay Baxter, Dan Lovell, Baxter Eaves, Vikash Mansinghka
#   Research Leads: Vikash Mansinghka, Patrick Shafto
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#

from pyparsing import *
## uses the module pyparsing. This document provides a very good explanation of pyparsing:
## http://www.nmt.edu/tcc/help/pubs/pyparsing/pyparsing.pdf

## Matches any white space and stores it as a single space
single_white = White().setParseAction(replaceWith(' '))

## basic keywords
and_keyword = CaselessKeyword("and")
from_keyword = CaselessKeyword("from")
for_keyword = CaselessKeyword("for")
into_keyword = CaselessKeyword("into")
of_keyword = CaselessKeyword("of")
with_keyword = CaselessKeyword("with")
help_keyword = CaselessKeyword("help").setResultsName("statement_id")
quit_keyword = CaselessKeyword("quit").setResultsName("statement_id")
to_keyword = CaselessKeyword('to')
## Many basic keywords will never be used alone
## creating them separately like this allows for simpler whitespace and case flexibility
conf_keyword = CaselessKeyword("conf")
create_keyword = CaselessKeyword("create")
upgrade_keyword = CaselessKeyword("upgrade")
execute_keyword = CaselessKeyword("execute")
file_keyword = CaselessKeyword("file")
update_keyword = CaselessKeyword("update")
metadata_keyword = CaselessKeyword('metadata')
label_keyword = CaselessKeyword('label')
codebook_keyword = CaselessKeyword('codebook')
schema_keyword = CaselessKeyword("schema")
set_keyword = CaselessKeyword("set")
categorical_keyword = CaselessKeyword("categorical")
numerical_keyword = CaselessKeyword("numerical")
cyclic_keyword = CaselessKeyword("cyclic")
ignore_keyword = CaselessKeyword("ignore")
key_keyword = CaselessKeyword("key")
wait_keyword = CaselessKeyword("wait")
cancel_keyword = CaselessKeyword("cancel")
initialize_keyword = CaselessKeyword("initialize").setResultsName("statement_id")
initialize_keyword.setParseAction(replaceWith("initialize_models"))
analyze_keyword = CaselessKeyword("analyze").setResultsName("statement_id")
index_keyword = CaselessKeyword("index")
save_keyword = CaselessKeyword("save")
load_keyword = CaselessKeyword("load")
drop_keyword = CaselessKeyword("drop")
show_keyword = CaselessKeyword("show")
select_keyword = CaselessKeyword("select")
infer_keyword = CaselessKeyword("infer")
simulate_keyword = CaselessKeyword("simulate")
estimate_keyword = CaselessKeyword("estimate")
pairwise_keyword = CaselessKeyword("pairwise")
where_keyword = CaselessKeyword("where")
order_keyword = CaselessKeyword("order")
by_keyword = CaselessKeyword("by")
limit_keyword = CaselessKeyword("limit")
confidence_keyword = CaselessKeyword("confidence")
times_keyword = CaselessKeyword("times")
dependence_keyword = CaselessKeyword("dependence")
probability_keyword = CaselessKeyword("probability")
correlation_keyword = CaselessKeyword("correlation")
mutual_keyword = CaselessKeyword("mutual")
information_keyword = CaselessKeyword("information")
typicality_keyword = CaselessKeyword("typicality")
as_keyword = CaselessKeyword("as")
show_keyword = CaselessKeyword("show")
similarity_keyword = CaselessKeyword("similarity")
respect_keyword = CaselessKeyword("respect")
predictive_keyword = CaselessKeyword("predictive")
group_keyword = CaselessKeyword("group")
diagnostics_keyword = CaselessKeyword("diagnostics")
summarize_keyword = CaselessKeyword("summarize").setResultsName("summarize")
freq_keyword = CaselessKeyword("freq").setResultsName("freq")
hist_keyword = CaselessKeyword("hist").setResultsName("hist")
plot_keyword = CaselessKeyword("plot").setResultsName("plot")
connected_keyword = CaselessKeyword("connected")
components_keyword = CaselessKeyword("components")
clusters_keyword = CaselessKeyword("clusters")
threshold_keyword = CaselessKeyword("threshold")
row_keyword = CaselessKeyword("row")
key_keyword = CaselessKeyword("key")
in_keyword = CaselessKeyword("in")
kernel_keyword = CaselessKeyword('kernel')
asc_keyword = CaselessKeyword("asc")
desc_keyword = CaselessKeyword("desc")
given_keyword = CaselessKeyword("given")
crp_keyword = CaselessKeyword("crp")
mixture_keyword = CaselessKeyword("mixture")
naive_keyword = CaselessKeyword("naive")
bayes_keyword = CaselessKeyword("bayes")
config_keyword = CaselessKeyword("config")
using_keyword = CaselessKeyword("using")
all_keyword = CaselessKeyword("all")
describe_keyword = CaselessKeyword("describe").setResultsName("statement_id")
## Single and plural keywords
single_model_keyword = CaselessKeyword("model")
multiple_models_keyword = CaselessKeyword("models")
single_iteration_keyword = CaselessKeyword("iteration")
multiple_iterations_keyword = CaselessKeyword("iterations")
single_sample_keyword = CaselessKeyword("sample")
multiple_samples_keyword = CaselessKeyword("samples")
single_column_keyword = CaselessKeyword("column")
multiple_columns_keyword = CaselessKeyword("columns")
single_list_keyword = CaselessKeyword("list")
multiple_lists_keyword = CaselessKeyword("lists")
single_btable_keyword = CaselessKeyword("btable")
multiple_btable_keyword = CaselessKeyword("btables")
single_minute_keyword = CaselessKeyword("minute")
multiple_minutes_keyword = CaselessKeyword("minutes")
single_description_keyword = CaselessKeyword("description")
multiple_descriptions_keyword = CaselessKeyword("descriptions")
single_short_name_keyword = CaselessKeyword("short") + single_white + CaselessKeyword("name")
multiple_short_names_keyword = CaselessKeyword("short") + single_white + CaselessKeyword("names")
## Plural agnostic syntax, setParseAction makes it all display the singular
description_keyword = single_description_keyword | multiple_descriptions_keyword
short_name_keyword = single_short_name_keyword | multiple_short_names_keyword
model_keyword = single_model_keyword | multiple_models_keyword
model_keyword.setParseAction(replaceWith("model"))
iteration_keyword = single_iteration_keyword | multiple_iterations_keyword
iteration_keyword.setParseAction(replaceWith("iteration"))
sample_keyword = single_sample_keyword | multiple_samples_keyword
sample_keyword.setParseAction(replaceWith("sample"))
column_keyword = single_column_keyword | multiple_columns_keyword
column_keyword.setParseAction(replaceWith("column"))
list_keyword = single_list_keyword | multiple_lists_keyword
list_keyword.setParseAction(replaceWith("list"))
btable_keyword = single_btable_keyword | multiple_btable_keyword
btable_keyword.setParseAction(replaceWith("btable"))
minute_keyword = single_minute_keyword | multiple_minutes_keyword
minute_keyword.setParseAction(replaceWith("minute"))

## Composite keywords: Inseparable elements that can have whitespace
## Using single_white and Combine to make them one string
execute_file_keyword = Combine(execute_keyword + single_white + file_keyword).setResultsName("statement_id")
execute_file_keyword.setParseAction(replaceWith("execute_file"))
create_btable_keyword = Combine(create_keyword + single_white + btable_keyword).setResultsName("statement_id")
create_btable_keyword.setParseAction(replaceWith('create_btable'))
upgrade_btable_keyword = Combine(upgrade_keyword + single_white + btable_keyword).setResultsName("statement_id")
upgrade_btable_keyword.setParseAction(replaceWith('upgrade_btable'))

update_codebook_for_keyword = Combine(update_keyword + single_white + codebook_keyword +
                                      single_white + for_keyword).setResultsName("statement_id")
update_codebook_for_keyword.setParseAction(replaceWith("update_codebook"))

update_descriptions_for_keyword = Combine(update_keyword + single_white + description_keyword +
                                single_white + for_keyword).setResultsName("statement_id")
update_descriptions_for_keyword.setParseAction(replaceWith("update_descriptions"))

update_short_names_for_keyword = Combine(update_keyword + single_white + short_name_keyword +
                               single_white + for_keyword).setResultsName("statement_id")
update_short_names_for_keyword.setParseAction(replaceWith("update_short_names"))

update_schema_for_keyword = Combine(update_keyword + single_white +
                                    schema_keyword + single_white + for_keyword).setResultsName("statement_id")
update_schema_for_keyword.setParseAction(replaceWith("update_schema"))
update_metadata_for_keyword = Combine(update_keyword + single_white +
                                    metadata_keyword + single_white + for_keyword).setResultsName("statement_id")
update_metadata_for_keyword.setParseAction(replaceWith("update_metadata"))
label_columns_for_keyword = Combine(label_keyword + single_white +
                                    column_keyword + single_white + for_keyword).setResultsName("statement_id")
label_columns_for_keyword.setParseAction(replaceWith("label_columns"))
show_columns_keyword = Combine(show_keyword + single_white + multiple_columns_keyword).setResultsName("statement_id")
show_columns_keyword.setParseAction(replaceWith("show_columns"))
show_metadata_for_keyword = Combine(show_keyword + single_white +
                                    metadata_keyword + single_white + for_keyword).setResultsName("statement_id")
show_metadata_for_keyword.setParseAction(replaceWith("show_metadata"))
show_label_for_keyword = Combine(show_keyword + single_white +
                                 label_keyword + single_white + for_keyword).setResultsName("statement_id")
show_label_for_keyword.setParseAction(replaceWith("show_label"))
show_analyze_for_keyword = Combine(show_keyword + single_white + analyze_keyword + single_white + for_keyword).setResultsName('statement_id')
show_analyze_for_keyword.setParseAction(replaceWith("show_analyze"))
cancel_analyze_for_keyword = Combine(cancel_keyword + single_white + analyze_keyword + single_white + for_keyword).setResultsName('statement_id')
cancel_analyze_for_keyword.setParseAction(replaceWith('cancel_analyze'))
models_for_keyword = Combine(model_keyword + single_white + for_keyword)
model_index_keyword = Combine(model_keyword + single_white + index_keyword)
load_model_keyword = Combine(load_keyword + single_white + model_keyword).setResultsName("statement_id")
load_model_keyword.setParseAction(replaceWith('load_models'))
save_model_keyword = Combine(save_keyword + single_white + model_keyword).setResultsName("statement_id")
save_model_keyword.setParseAction(replaceWith("save_models"))
save_to_keyword = Combine(save_keyword + single_white + to_keyword).setResultsName("statement_id")
list_btables_keyword = Combine(list_keyword + single_white + btable_keyword).setResultsName("statement_id")
list_btables_keyword.setParseAction(replaceWith("list_btables"))
drop_btable_keyword = Combine(drop_keyword + single_white + btable_keyword).setResultsName("statement_id")
drop_btable_keyword.setParseAction(replaceWith("drop_btable"))
drop_column_list_keyword = Combine(drop_keyword + single_white + column_keyword + single_white + list_keyword).setResultsName("statement_id")
drop_column_list_keyword.setParseAction(replaceWith("drop_column_list"))
drop_row_list_keyword = Combine(drop_keyword + single_white + row_keyword + single_white + list_keyword).setResultsName("statement_id")
drop_row_list_keyword.setParseAction(replaceWith("drop_row_list"))
drop_model_keyword = Combine(drop_keyword + single_white + model_keyword).setResultsName("statement_id")
drop_model_keyword.setParseAction(replaceWith("drop_models"))
show_schema_for_keyword = Combine(show_keyword + single_white + schema_keyword +
                                  single_white + for_keyword).setResultsName("statement_id")
show_schema_for_keyword.setParseAction(replaceWith("show_schema"))
show_models_for_keyword = Combine(show_keyword + single_white + model_keyword +
                                  single_white + for_keyword).setResultsName("statement_id")
show_models_for_keyword.setParseAction(replaceWith("show_models"))
show_diagnostics_for_keyword = Combine(show_keyword + single_white + diagnostics_keyword +
                                       single_white + for_keyword).setResultsName("statement_id")
show_diagnostics_for_keyword.setParseAction(replaceWith("show_diagnostics"))
show_column_lists_for_keyword = Combine(show_keyword + single_white + column_keyword +
                                        single_white + list_keyword +
                                        single_white + for_keyword).setResultsName("statement_id")
show_column_lists_for_keyword.setParseAction(replaceWith("show_column_lists"))
show_row_lists_for_keyword = Combine(show_keyword + single_white + row_keyword +
                                 single_white + list_keyword +
                                 single_white + for_keyword).setResultsName("statement_id")
show_row_lists_for_keyword.setParseAction(replaceWith("show_row_lists"))
estimate_pairwise_keyword = Combine(estimate_keyword + single_white +
                                    pairwise_keyword).setResultsName("statement_id")
estimate_pairwise_keyword.setParseAction(replaceWith("estimate_pairwise"))
create_column_list_keyword = Combine(create_keyword + single_white + single_column_keyword + single_white + single_list_keyword).setResultsName("statement_id")
create_column_list_keyword.setParseAction(replaceWith("create_column_list"))
estimate_pairwise_row_keyword = Combine(estimate_keyword + single_white + pairwise_keyword +
                                        single_white + row_keyword).setResultsName("statement_id")
estimate_pairwise_row_keyword.setParseAction(replaceWith('estimate_pairwise_row'))
row_similarity_keyword = Combine(row_keyword + single_white + similarity_keyword)
with_confidence_keyword = Combine(with_keyword + single_white + confidence_keyword)
order_by_keyword = Combine(order_keyword + single_white + by_keyword)
dependence_probability_keyword = Combine(dependence_keyword + single_white + probability_keyword)
mutual_information_keyword = Combine(mutual_keyword + single_white + information_keyword)
estimate_columns_from_keyword = Combine(estimate_keyword + single_white + column_keyword +
                                        single_white + from_keyword).setResultsName("statement_id")
estimate_columns_keyword = Combine(estimate_keyword + single_white + column_keyword).setResultsName("statement_id")
estimate_columns_from_keyword.setParseAction(replaceWith("estimate_columns"))
column_lists_keyword = Combine(column_keyword + single_white + list_keyword)
similarity_to_keyword = Combine(similarity_keyword + single_white + to_keyword).setResultsName("statement_id")
with_respect_to_keyword = Combine(with_keyword + single_white + respect_keyword + single_white + to_keyword)
probability_of_keyword = Combine(probability_keyword + single_white + of_keyword)
probability_of_keyword.setParseAction(replaceWith("probability"))
typicality_of_keyword = Combine(typicality_keyword + single_white + of_keyword)
predictive_probability_of_keyword = Combine(predictive_keyword + single_white + probability_keyword + single_white + of_keyword)
predictive_probability_of_keyword.setParseAction(replaceWith("predictive probability"))

save_connected_components_with_threshold_keyword = Combine(save_keyword + single_white +
                                                           connected_keyword + single_white +
                                                           components_keyword + single_white +
                                                           with_keyword + single_white +
                                                           threshold_keyword)
save_connected_components_keyword = save_connected_components_with_threshold_keyword

save_clusters_with_threshold_keyword = Combine(save_keyword + single_white +
                                                clusters_keyword + single_white +
                                                with_keyword + single_white +
                                                threshold_keyword)
save_clusters_keyword = save_clusters_with_threshold_keyword

key_in_keyword = Combine(key_keyword + single_white + in_keyword)
naive_bayes_keyword = Combine(naive_keyword + single_white + bayes_keyword)
crp_mixture_keyword = Combine(crp_keyword + single_white + mixture_keyword)
using_models_keyword = Combine(using_keyword + single_white + model_keyword)

## Values/Literals
sub_query = QuotedString("(",endQuoteChar=")").setResultsName('sub_query')
float_number = Regex(r'[-+]?[0-9]*\.?[0-9]+') | sub_query
int_number = Word(nums) | sub_query
operation_literal = oneOf("<= >= = < >")
equal_literal = Literal("=")
semicolon_literal = Literal(";")
comma_literal = Literal(",")
hyphen_literal = Literal("-")
paren_open_literal = Literal("(")
paren_close_literal = Literal(")")
all_column_literal = Literal('*')
identifier = (Word(alphas + '/', alphanums + "_./").setParseAction(downcaseTokens)) | sub_query
btable = identifier.setResultsName("btable") | sub_query
comment = (Literal('--') + restOfLine).suppress()
# single and double quotes inside value must be escaped.
value = (QuotedString('"', escChar='\\') |
         QuotedString("'", escChar='\\') |
         Word(printables,excludeChars=',;')|
         float_number |
         sub_query)
filename = (QuotedString('"', escChar='\\') |
            QuotedString("'", escChar='\\') |
            Word(alphanums + "!\"/#$%&'()*+,-.:<=>?@[\]^_`{|}~")).setResultsName("filename")
label = (QuotedString('"', escChar='\\') |
         QuotedString("'", escChar='\\') |
         Word(printables, excludeChars=',;'))
data_type_literal = categorical_keyword | numerical_keyword | ignore_keyword | key_keyword | cyclic_keyword

cyclic_parameters = Group(
    Suppress(paren_open_literal) +
    float_number.setResultsName("min") +
    Suppress(comma_literal) +
    float_number.setResultsName("max") +
    Suppress(paren_close_literal)).setResultsName("parameters")

categorical_parameters = Group(
    Suppress(paren_open_literal) +
    int_number.setResultsName("cardinality") +
    Suppress(paren_close_literal)).setResultsName("parameters")

cyclic = cyclic_keyword + cyclic_parameters
categorical = categorical_keyword + Optional(categorical_parameters)

data_type = categorical | numerical_keyword | ignore_keyword | key_keyword | cyclic


###################################################################################
# ------------------------------------ Functions -------------------------------- #
###################################################################################

# ------------------------------- Management statements ------------------------- #

# CREATE BTABLE <btable> FROM <filename.csv> [WITH CODEBOOK <codebook.csv>]
create_btable_function = create_btable_keyword + btable + Suppress(from_keyword) + filename + Optional(
   with_keyword + codebook_keyword + filename.setResultsName('codebook_file'))

# UPGRADE BTABLE <btable>
upgrade_btable_function = upgrade_btable_keyword + btable

# UPDATE SCHEMA FOR <btable> SET <col1>=<type1>[,<col2>=<type2>...]


type_clause = Group(
    ZeroOrMore(
      Group(identifier + Suppress(equal_literal) + data_type) + Suppress(comma_literal)) +
      Group(identifier + Suppress(equal_literal) + data_type)).setResultsName("type_clause")

update_schema_for_function = (update_schema_for_keyword +
                              btable +
                              Suppress(set_keyword) +
                              type_clause)


label_clause = Group(ZeroOrMore(Group(identifier + Suppress(equal_literal) + OneOrMore(label)) +
                                Suppress(comma_literal)) +
                                Group(identifier + Suppress(equal_literal) + OneOrMore(label))).setResultsName("label_clause")

# UPDATE CODEBOOK FOR <btable> FROM <filename.csv>
update_codebook_for_function = (update_codebook_for_keyword + btable + Suppress(from_keyword) +
                                filename)

# UPDATE DESCRIPTIONS FOR <btable> SET <column1 = column-desc-1> [, <column-name-2 = column-desc-2>, ...]
update_descriptions_for_function = (update_descriptions_for_keyword + btable +
                                    Suppress(set_keyword) + label_clause)

# UPDATE SHORT NAMES FOR <btable> SET <column1 = column-desc-1> [, <column-name-2 = column-desc-2>, ...]
update_short_names_for_function = (update_short_names_for_keyword + btable +
                                   Suppress(set_keyword) + label_clause)

# UPDATE METADATA FOR <btable> (SET <metadata-key1 = value1>[, <metadata-key2 = value2>...] | FROM <filename.csv>)
update_metadata_for_function = (update_metadata_for_keyword + btable +
                                (set_keyword + label_clause | from_keyword + filename))

# LABEL COLUMNS FOR <btable> (SET <column1 = column-label-1> [, <column-name-2 = column-label-2>, ...] | FROM <filename.csv>)
label_columns_for_function = (label_columns_for_keyword + btable +
                              (set_keyword + label_clause | from_keyword + filename))

# SHOW COLUMNS <column_list> FOR <btable>
show_columns_function = (show_columns_keyword +
                         Optional(identifier.setResultsName('column_list')) +
                         for_keyword +
                         btable)

# SHOW METADATA FOR <btable> [<metadata-key1> [, <metadata-key2>...]]
show_metadata_function = (show_metadata_for_keyword +
                          btable +
                          Optional(Group(identifier +
                                         ZeroOrMore(Suppress(comma_literal) + identifier)).setResultsName('keyset')))

# SHOW LABEL FOR <btable> [<column-name-1> [, <column-name-2>...]]
show_label_function = (show_label_for_keyword +
                       btable +
                       Optional(Group(identifier +
                                      ZeroOrMore(Suppress(comma_literal) + identifier)).setResultsName('columnset')))

# EXECUTE FILE <filename.bql>
execute_file_function = execute_file_keyword + filename

# INITIALIZE <num_models> MODELS FOR <btable>
initialize_function = (initialize_keyword +
                       Optional(int_number.setResultsName("num_models")) +
                       Suppress(models_for_keyword) +
                       btable +
                       Optional(with_keyword +
                                config_keyword +
                                (naive_bayes_keyword|crp_mixture_keyword).setResultsName('config')))

# ANALYZE <btable> [MODEL[S] <model_index>-<model_index>] FOR (<num_iterations> ITERATIONS | <minutes> MINUTES)
def list_from_index_clause(toks):
    ## takes index tokens separated by '-' for range and ',' for individual and returns a list of unique indexes
    index_list = []
    for token in toks[0]:
        if type(token)== str:
            index_list.append(int(token))
        elif len(token) == 2:
            index_list += range(int(token[0]), int(token[1])+1)
        else:
            pass
            ##TODO throw exception
        index_list.sort()
    return [list(set(index_list))]

index_clause = (Group((Group(int_number + Suppress(hyphen_literal) + int_number) | int_number) +
                      ZeroOrMore(Suppress(comma_literal) +
                                 (Group(int_number + Suppress(hyphen_literal) + int_number) |
                                  int_number)))
                .setParseAction(list_from_index_clause)
                .setResultsName('index_clause'))

with_kernel_clause = Group(with_keyword + identifier.setResultsName('kernel_id') + kernel_keyword).setResultsName("with_kernel_clause")

analyze_function = (analyze_keyword + btable +
                    Optional( model_keyword + index_clause) +
                    Suppress(for_keyword) +
                    ((int_number.setResultsName('num_iterations') + iteration_keyword) |
                     (int_number.setResultsName('num_minutes') + minute_keyword)) + Optional(with_kernel_clause) +
                    Optional(wait_keyword).setResultsName('wait'))

# LIST BTABLES
list_btables_function = list_btables_keyword

show_for_btable_statement = ((show_schema_for_keyword |
                              show_models_for_keyword |
                              show_diagnostics_for_keyword |
                              show_column_lists_for_keyword |
                              show_analyze_for_keyword |
                              show_row_lists_for_keyword) +
                             btable)
cancel_analyze_for_function = cancel_analyze_for_keyword + btable

# LOAD MODELS <filename.pkl.gz> INTO <btable>
load_model_function = load_model_keyword + filename + Suppress(into_keyword) + btable

# SAVE MODELS FROM <btable> TO <filename.pkl.gz>
save_model_from_function = save_model_keyword + Suppress(from_keyword) + btable + Suppress(to_keyword) + filename

# DROP BTABLE <btable>
drop_btable_function = drop_btable_keyword + btable

# DROP COLUMN LIST <listname> FROM <btable>
drop_column_list_function = drop_column_list_keyword + identifier.setResultsName('list_name') + Suppress(from_keyword) + btable

# DROP ROW LIST <listname> FROM <btable>
drop_row_list_function = drop_row_list_keyword + identifier.setResultsName('list_name') + Suppress(from_keyword) + btable

# DROP MODEL[S] [<model_index>-<model_index>] FROM <btable>
drop_model_function = drop_model_keyword + Optional(index_clause) + Suppress(from_keyword) + btable

# DESCRIBE [<column-name-1> [, <column-name-2>...]] FOR <btable>
describe_function = describe_keyword + Optional(Group(identifier +
               (ZeroOrMore(Suppress(comma_literal) + identifier)) | all_column_literal).setResultsName('columnset')) + Suppress(for_keyword) + btable

# Help [function name]
help_function = help_keyword + Optional(Word(alphas).setParseAction(downcaseTokens)).setResultsName("method_name") + ZeroOrMore(Word(alphas))
quit_function = quit_keyword

management_query = (create_btable_function |
                    upgrade_btable_function |
                    update_schema_for_function |
                    execute_file_function |
                    initialize_function |
                    analyze_function |
                    list_btables_function |
                    show_for_btable_statement |
                    load_model_function |
                    save_model_from_function |
                    drop_btable_function |
                    drop_column_list_function |
                    drop_row_list_function |
                    drop_model_function |
                    help_function |
                    update_metadata_for_function |
                    label_columns_for_function |
                    show_label_function |
                    describe_function |
                    update_descriptions_for_function |
                    update_short_names_for_function |
                    update_codebook_for_function |
                    show_metadata_function |
                    show_columns_function |
                    cancel_analyze_for_function |
                    quit_function)

# ------------------------------ Helper Clauses --------------------------- #

# Rows can be identified either by an integer or <column> = <value> where value is unique for the given column
row_clause = (int_number.setResultsName("row_id") |
              (identifier.setResultsName("column") +
               equal_literal +
               value.setResultsName("column_value")))

column_list_clause = Group((identifier | all_column_literal) +
                           ZeroOrMore(Suppress(comma_literal|and_keyword) +
                                      (identifier | all_column_literal))).setResultsName("column_list")

# SAVE TO <file>
save_to_clause = save_to_keyword + filename

# WITH CONFIDENCE <confidence>
with_confidence_clause = with_confidence_keyword + float_number.setResultsName("confidence")

# -------------------------------- Functions ------------------------------ #

# SIMILARITY TO <row> [WITH RESPECT TO <column>]
similarity_to_function = (Group(similarity_keyword.setResultsName('function_id') +
                                Optional(to_keyword.setResultsName('to') +
                                         row_clause) +
                                Optional(with_respect_to_keyword + column_list_clause)
                                .setResultsName('with_respect_to'))
                          .setResultsName("function"))

# TYPICALITY [OF <column>]
typicality_function = Group(typicality_keyword.setResultsName('function_id') + Optional(of_keyword + identifier.setResultsName("column"))).setResultsName("function")

# Functions of two columns for use in dependence probability, mutual information, correlation
functions_of_two_columns_subclause = ((with_keyword +
                                       identifier.setResultsName("with_column")) |
                                      (of_keyword +
                                       identifier.setResultsName("of_column") +
                                       with_keyword +
                                       identifier.setResultsName("with_column")))

##TODO make all of these functions one thing
# DEPENDENCE PROBABILITY (WITH <column> | OF <column1> WITH <column2>)
dependence_probability_function = (Group(dependence_probability_keyword.setResultsName('function_id') +
                                        Optional(functions_of_two_columns_subclause)).setResultsName("function"))

# MUTUAL INFORMATION [OF <column1> WITH <column2>]
mutual_information_function = (Group(mutual_information_keyword.setResultsName('function_id') +
                                    Optional(functions_of_two_columns_subclause)).setResultsName("function"))

# CORRELATION [OF <column1> WITH <column2>]
correlation_function = (Group(correlation_keyword.setResultsName('function_id') +
                             Optional(functions_of_two_columns_subclause)).setResultsName("function"))

two_column_function = (dependence_probability_function | mutual_information_function | correlation_function)

# PROBABILITY OF <column>=<value>
probability_of_function = Group((probability_of_keyword.setResultsName("function_id") +
                                 identifier.setResultsName("column") +
                                 equal_literal.setResultsName('op') +
                                 value.setResultsName("value"))).setResultsName('function')

# PREDICTIVE PROBABILITY OF <column>
predictive_probability_of_function = Group(predictive_probability_of_keyword.setResultsName("function_id") +
                                           identifier.setResultsName("column")).setResultsName("function")

# KEY IN <row_list>
key_in_rowlist_clause = Group(key_keyword.setResultsName("function_id")).setResultsName('function') + in_keyword.setResultsName('operation') + identifier.setResultsName("value")

whereclause_potential_function = (similarity_to_function |
                                  typicality_function |
                                  predictive_probability_of_function |
                                  two_column_function |
                                  Group(identifier.setResultsName('column')))

# -------------------------------- other clauses --------------------------- #

# ORDER BY <column|non-aggregate-function>[<column|function>...] [asc|desc]
single_order_by = Group(whereclause_potential_function.setResultsName('function') +
                        Optional(asc_keyword | desc_keyword).setResultsName('asc_desc') +
                        Optional(conf_keyword + float_number.setResultsName('conf')))

order_by_clause = Group(Suppress(order_by_keyword) +
                        single_order_by +
                        ZeroOrMore(Suppress(comma_literal) + single_order_by)).setResultsName('order_by')

# WHERE <whereclause>
single_where_condition = Group(((whereclause_potential_function.setResultsName('function') +
                                 operation_literal.setResultsName('operation') +
                                 value.setResultsName('value')) | key_in_rowlist_clause) +
                               Optional(conf_keyword + float_number.setResultsName('conf')))

where_clause = (where_keyword.setResultsName('where_keyword') +
                Group(single_where_condition +
                      ZeroOrMore(Suppress(and_keyword) + single_where_condition))
                .setResultsName("where_conditions"))

save_connected_components_clause = Group(save_connected_components_keyword
                                         .setResultsName("save_connected_components") +
                                         float_number.setResultsName("threshold") +
                                         ((as_keyword +
                                           identifier.setResultsName('as_label')) |
                                          (into_keyword +
                                           identifier.setResultsName('into_label')))).setResultsName('connected_components_clause')

save_clusters_clause = Group(save_clusters_keyword
                            .setResultsName('save_clusters') +
                            float_number.setResultsName('threshold') +
                            ((as_keyword +
                              identifier.setResultsName('as_label')) |
                             (into_keyword +
                              identifier.setResultsName('into_label')))).setResultsName('clusters_clause')

list_clause = Group(int_number|identifier +
                    ZeroOrMore(Suppress(comma_literal) +
                               (int_number|identifier)))

single_given_condition = Group(identifier.setResultsName('column') + equal_literal + value.setResultsName('value'))
given_clause = (Group(given_keyword +
                      Group(single_given_condition +
                            ZeroOrMore(Suppress(comma_literal|and_keyword) +
                                       single_given_condition))
                      .setResultsName("given_conditions"))
                .setResultsName("given_clause"))

using_models_clause = (using_models_keyword + index_clause.setResultsName("using_models_index_clause"))

# ----------------------------- Master Query Syntax ---------------------------------------- #

query_id = (select_keyword |
            infer_keyword |
            simulate_keyword |
            estimate_pairwise_row_keyword |
            estimate_pairwise_keyword |
            estimate_keyword |
            create_column_list_keyword).setResultsName('statement_id')

function_in_query = (predictive_probability_of_function |
                     probability_of_function |
                     typicality_function |
                     similarity_to_function |
                     dependence_probability_function |
                     mutual_information_function |
                     correlation_function |
                     column_keyword |
                     dependence_probability_keyword |
                     Group((identifier|all_column_literal).setResultsName("column_id") +
                           Optional(conf_keyword + float_number.setResultsName('conf')))).setResultsName("function")

functions_clause = Group(function_in_query +
                         ZeroOrMore(Suppress(comma_literal) +
                                    function_in_query)).setResultsName('functions')

query = (Optional(freq_keyword | hist_keyword | summarize_keyword | plot_keyword) +
         query_id +
         functions_clause +
         Suppress(from_keyword) +
         btable +
         Optional(where_clause) +
         Each(Optional(order_by_clause) +
              Optional(with_confidence_clause) +
              Optional(Suppress(with_keyword) + int_number.setResultsName('samples') + Suppress(sample_keyword)) +
              Optional(Suppress(limit_keyword) + int_number.setResultsName("limit")) +
              Optional(given_clause) +
              Optional(for_keyword + list_clause.setResultsName('for_list')) +
              Optional(Suppress(save_to_keyword) + filename) +
              Optional(save_clusters_clause) +
              Optional(Suppress(times_keyword) + int_number.setResultsName("times")) +
              Optional(using_models_clause) +
              Optional(Suppress(as_keyword) + identifier.setResultsName("as_column_list")) +
              Optional(Suppress(into_keyword) + identifier.setResultsName("newtablename"))))

bql_statement = (query | management_query) + Optional(semicolon_literal)
bql_input = OneOrMore(Group(bql_statement))
bql_input.ignore(comment)



#
#   Copyright (c) 2010-2014, MIT Probabilistic Computing Project
#
#   Lead Developers: Jay Baxter and Dan Lovell
#   Authors: Jay Baxter, Dan Lovell, Baxter Eaves, Vikash Mansinghka
#   Research Leads: Vikash Mansinghka, Patrick Shafto
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#

from bayesdb.client import Client
from cmd2 import Cmd
import sys

class BayesDBApp(Cmd):
  """Provides "interactive mode" features."""
  Cmd.redirector = '>>>'

  def __init__(self, client):
    self.client = client
    self.prompt = 'bql> '
    Cmd.__init__(self, 'tab')


  def do_show(self, line):
    self.client('show ' + str(line))

  def do_list(self, line):
    self.client('list ' + str(line))

  def do_analyze(self, line):
    self.client('analyze ' + str(line))

  def do_execute(self, line):
    self.client('execute ' + str(line))

  def do_drop(self, line):
    self.client('drop ' + str(line))

  def do_initialize(self, line):
    self.client('initialize ' + str(line))

  def do_create(self, line):
    self.client('create ' + str(line))

  def do_infer(self, line):
    self.client('infer ' + str(line))

  def do_select(self, line):
    self.client('select ' + str(line))

  def do_simulate(self, line):
    self.client('simulate ' + str(line))

  def do_save(self, line):
    self.client('save ' + str(line))
    
  def do_load(self, line):
    self.client('load ' + str(line))

  def do_estimate(self, line):
    self.client('estimate ' + str(line))

  def do_update(self, line):
    self.client('update ' + str(line))

  def do_help(self, line):
    self.client('help ' + str(line))
    
  def default(self, line):
    self.client(str(line))

def run_command_line():
  # Get command line arguments to specify hostname and port
  hostname = None
  port = None
  if len(sys.argv) > 1:
    # Treat the first argument as hostname[:port]
    input = sys.argv[1].split(':')
    hostname = input[0]
    if len(input) == 1:
      client = Client(hostname)
      print "Using hostname %s." % hostname
    if len(input) == 2:
      port = int(input[1])
      client = Client(hostname, port)
      print "Using hostname %s, port %d" % (hostname, port)
    elif len(input) > 2:
      print "Run with 'python bql [hostname[:port]]'"
  else:
    client = Client()

  print """Welcome to BayesDB. You may enter BQL commands directly into this prompt. Type 'help' for help, and 'quit' to quit."""
  app = BayesDBApp(client)
  app.cmdloop()

if __name__ == "__main__":
  run_command_line()



#
#   Copyright (c) 2010-2014, MIT Probabilistic Computing Project
#
#   Lead Developers: Jay Baxter and Dan Lovell
#   Authors: Jay Baxter, Dan Lovell, Baxter Eaves, Vikash Mansinghka
#   Research Leads: Vikash Mansinghka, Patrick Shafto
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#

import os 
import sys 
import pytest

from bayesdb.client import Client

def teardown_function(function):
    for fname in os.listdir('.'):
        if fname[-4] == '.png':
            os.remove(fname)

def run_example(name):
    # Default upgrade_key_column is None, to let the user choose, but need to avoid
    # user input during testing, so default will be to create a new key column.
    client = Client(testing=True)
    file_path = os.path.join('../../examples/%s/%s_analysis.bql' % (name, name))
    results = client(open(file_path, 'r'), yes=True, pretty=False, plots=False, key_column=0)
    for r in results:
        if 'Error' in r or ('error' in r and r['error']):
            raise Exception(str(r))

def test_dha_example():
    run_example('dha')

def test_gss_example():
    run_example('gss')
    
def test_chicago_small_example():
    run_example('chicago_small')

def test_flights_example():
    run_example('flights')

def test_kiva_example():
    run_example('kiva')

def test_employees_example():
    run_example('employees')
    
    



#
#   Copyright (c) 2010-2014, MIT Probabilistic Computing Project
#
#   Lead Developers: Jay Baxter and Dan Lovell
#   Authors: Jay Baxter, Dan Lovell, Baxter Eaves, Vikash Mansinghka
#   Research Leads: Vikash Mansinghka, Patrick Shafto
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#

import numpy as np

# there should be a 'get' method for each item in diagnosics lists
multi_state_diagnostics = []
single_state_diagnostics = [
    'num_views',
    'column_crp_alpha',
    'min_clusters_view',
    'max_clusters_view',
    'mean_clusters_view',
    'std_clusters_view'
]

# muti-state diagnostics

# single-state methods
def get_num_views(X_L, X_D):
    """Returns the number of views in X_L
    """
    return len(X_L['column_partition']['counts'])


def get_column_crp_alpha(X_L, X_D):
    """Retruns column crp alpha parameter
    """
    return X_L['column_partition']['hypers']['alpha']


def get_min_clusters_view(X_L, X_D):
    """ Returns the minimum number of clusters in a view
    """
    Z = np.array(X_D, dtype=int)
    num_clusters_view = np.max(Z, axis=1)+1
    return np.min(num_clusters_view)


def get_max_clusters_view(X_L, X_D):
    """ Returns the maximum number of clusters in a view
    """
    Z = np.array(X_D, dtype=int)
    num_clusters_view = np.max(Z, axis=1)+1
    return np.max(num_clusters_view)


def get_mean_clusters_view(X_L, X_D):
    """ Returns the mean number of clusters in views
    """
    Z = np.array(X_D, dtype=int)
    num_clusters_view = np.max(Z, axis=1)+1
    return np.mean(num_clusters_view)


def get_std_clusters_view(X_L, X_D):
    """ Returns the standard deviation of numbers of clusters in views
    """
    Z = np.array(X_D, dtype=int)
    num_clusters_view = np.max(Z, axis=1)+1
    return np.std(num_clusters_view)



#
#   Copyright (c) 2010-2014, MIT Probabilistic Computing Project
#
#   Lead Developers: Jay Baxter and Dan Lovell
#   Authors: Jay Baxter, Dan Lovell, Baxter Eaves, Vikash Mansinghka
#   Research Leads: Vikash Mansinghka, Patrick Shafto
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#

import re
import utils
import numpy
import os
import pylab
import matplotlib.cm
import inspect
import operator
import ast
import string

import utils
import functions
import data_utils as du
from pyparsing import *
import bayesdb.bql_grammar as bql_grammar

def evaluate_where_on_row(row_idx, row, where_conditions, M_c, M_c_full, X_L_list, X_D_list, T, T_full, engine, tablename, numsamples, impute_confidence):
  """
  Helper function that applies WHERE conditions to row, returning False if row doesn't satisfy where
  clause, and the list of function results if it does satisfy the where clause.
  """
  function_values = []
  for (func, f_args, op, val) in where_conditions:
    if func == functions._column and f_args[1] != None and numpy.isnan(T[row_idx][f_args[0]]):
      col_idx = f_args[0]
      confidence = f_args[1]
      ## need to do predictive sampling to evaluate where condition with confidence
      ## TODO: easier way to do this would be to call impute on backend, but would need to change
      ## crosscat so that impute_and_confidence could also return the original samples, or evaluate
      ## a whereclause.
      Y = [(row_idx, cidx, row[cidx]) for cidx in M_c['name_to_idx'].values() \
           if not numpy.isnan(T[row_idx][cidx])]      
      samples = engine.call_backend('simple_predictive_sample',
                   dict(M_c=M_c, X_L=X_L_list, X_D=X_D_list, Y=Y, Q=[[row_idx,col_idx]], n=numsamples))
      samples_satisfying_where = 0
      for sample in samples:
        value = du.convert_code_to_value(M_c, col_idx, sample[0])
        if op(value, val):
          samples_satisfying_where += 1
      if float(samples_satisfying_where)/len(samples) >= confidence:
        # Where clause is satisfied! Now, generate impute summary.
        imputed_code, imputation_confidence = utils.get_imputation_and_confidence_from_samples(
          M_c, X_L_list[0], col_idx, samples)
        if imputed_code is not None:
          imputed_value = du.convert_code_to_value(M_c, col_idx, imputed_code)
        else:
          imputed_value = T[row_idx][col_idx]
        function_values.append(imputed_value)
      else:
        return False
    else:
      if func != functions._column_ignore:
        where_value = func(f_args, row_idx, row, M_c, X_L_list, X_D_list, T, engine, numsamples)
      else:
        where_value = func(f_args, row_idx, row, M_c_full, T_full, engine)
      if func == functions._row_id:
        # val should be a row list name in this case. look up the row list, and set val to be the list of
        # row indices in the row list. Throws BayesDBRowListDoesNotExistError if row list does not exist.
        val = engine.persistence_layer.get_row_list(tablename, val)
        if op(val, where_value): # for operator.contains, op(a,b) means 'b in a': so need to switch args.
          function_values.append(where_value)
        else:
          return False
      else:
        # Normal, most common condition.
        if op(where_value, val):
          function_values.append(where_value)
        else:
          return False
  return function_values

def convert_row_from_codes_to_values(row, M_c):
  """
  Helper function to convert a row from its 'code' (as it's stored in T) to its 'value'
  (the human-understandable value).
  """
  ret = []
  for cidx, code in enumerate(row): 
    if not du.flexible_isnan(code):
      ret.append(du.convert_code_to_value(M_c, cidx, code))
    else:
      ret.append(code)
  return tuple(ret)

def check_if_functions_need_models(queries, tablename, order_by, where_conditions):
  """
  If there are no models, make sure that we aren't using functions that require models.
  TODO: make this less hardcoded
  """
  blacklisted_functions = [functions._similarity, functions._row_typicality, functions._col_typicality, functions._probability]
  used_functions = [q[0] for q in queries] + [w[0] for w in where_conditions] + [x[0] for x in order_by]
  for bf in blacklisted_functions:
    if bf in queries:
      raise utils.BayesDBNoModelsError(tablename)


def compute_result_and_limit(rows, limit, queries, M_c, X_L_list, X_D_list, T, engine, numsamples):
  data = []
  row_count = 0

  # Compute aggregate functions just once, then cache them.
  aggregate_cache = dict()
  for query_idx, (query_function, query_args, aggregate) in enumerate(queries):
    if aggregate:
      aggregate_cache[query_idx] = query_function(query_args, None, None, M_c, X_L_list, X_D_list, T, engine, numsamples)

  # Only return one row if all aggregate functions (row_id will never be aggregate, so subtract 1 and don't return it).
  assert queries[0][0] == functions._row_id
  if len(aggregate_cache) == len(queries) - 1:
    limit = 1

  # Iterate through data table, calling each query_function to fill in the output values.
  for row_id, row_values in rows:
    ret_row = []
    for query_idx, (query_function, query_args, aggregate) in enumerate(queries):
      if aggregate:
        ret_row.append(aggregate_cache[query_idx])
      else:
        ret_row.append(query_function(query_args, row_id, row_values, M_c, X_L_list, X_D_list, T, engine, numsamples))
    data.append(tuple(ret_row))
    row_count += 1
    if row_count >= limit:
      break
  return data



#
#   Copyright (c) 2010-2014, MIT Probabilistic Computing Project
#
#   Lead Developers: Jay Baxter and Dan Lovell
#   Authors: Jay Baxter, Dan Lovell, Baxter Eaves, Vikash Mansinghka
#   Research Leads: Vikash Mansinghka, Patrick Shafto
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#

import re
import utils
import numpy
import os
import pylab
import matplotlib.cm
import inspect
import operator
import ast

import utils
import functions
import data_utils as du


def filter_column_indices(column_indices, where_conditions, M_c, T, X_L_list, X_D_list, engine, numsamples):
  # should return tuples of the form (data, c_idx), where data is a list of the values of each of the where functions.
  column_indices_and_data = []

  for c_idx in column_indices:
    data = []
    
    for ((func, f_args), op, val) in where_conditions:
      # mutual_info, correlation, and dep_prob all take args=(i,j)
      # col_typicality takes just args=i
      # incoming f_args will be None for col_typicality, j for the three others
      if f_args is not None:
        f_args = (f_args, c_idx)
      else:
        f_args = c_idx
      where_value = func(f_args, None, None, M_c, X_L_list, X_D_list, T, engine, numsamples)
      data.append(where_value)
      if not op(where_value, val):
        break
        
    column_indices_and_data.append((data, c_idx))
    
  return column_indices_and_data



  
  return [c_idx for c_idx in column_indices if _is_column_valid(c_idx, where_conditions, M_c, X_L_list, X_D_list, T, engine, numsamples)]

def _is_column_valid(c_idx, where_conditions, M_c, X_L_list, X_D_list, T, engine, numsamples):
  for ((func, f_args), op, val) in where_conditions:
    # mutual_info, correlation, and dep_prob all take args=(i,j)
    # col_typicality takes just args=i
    # incoming f_args will be None for col_typicality, j for the three others
    if f_args is not None:
      f_args = (f_args, c_idx)
    else:
      f_args = c_idx
    where_value = func(f_args, None, None, M_c, X_L_list, X_D_list, T, engine, numsamples)
    return op(where_value, val)
  return True


def order_columns(column_indices, order_by, M_c, X_L_list, X_D_list, T, engine, numsamples):
  if not order_by:
    return column_indices
  return _column_order_by(column_indices, order_by, M_c, X_L_list, X_D_list, T, engine, numsamples)


def _column_order_by(column_indices_and_data, function_list, M_c, X_L_list, X_D_list, T, engine, numsamples):
  """
  Return the original column indices, but sorted by the individual functions.
  """
  if len(column_indices_and_data) == 0 or not function_list:
    return column_indices_and_data

  scored_column_indices = list() ## Entries are (score, cidx, data)
  for data, c_idx in column_indices_and_data:
    ## Apply each function to each cidx to get a #functions-length tuple of scores.
    scores = []
    values = []
    for (f, f_args, desc) in function_list:

      # mutual_info, correlation, and dep_prob all take args=(i,j)
      # col_typicality takes just args=i
      # incoming f_args will be None for col_typicality, j for the three others
      if f_args:
        f_args = (f_args, c_idx)
      else:
        f_args = c_idx

      score = f(f_args, None, None, M_c, X_L_list, X_D_list, T, engine, numsamples)
      data.append(score)
      # nan values create really unpredictable sort behavior, so set score to inf for consistency
      if numpy.isnan(score):
        score = float('inf')
      elif desc:
        score *= -1
      scores.append(score)
    scored_column_indices.append((tuple(scores), c_idx, tuple(data)))
  scored_column_indices.sort(key=lambda tup: tup[0], reverse=False)

  return [(data, c_idx) for (scores, c_idx, data) in scored_column_indices]

def function_description(func, f_args, M_c):
  function_names = {'_col_typicality': 'typicality',
    '_dependence_probability': 'dependence probability',
    '_correlation': 'correlation',
    '_mutual_information': 'mutual information'
    }

  function_name = function_names[func.__name__]

  if function_name == 'typicality':
    description = 'typicality'
  elif f_args is not None:
    function_arg = M_c['idx_to_name'][str(f_args)]
    description = '%s with %s' % (function_name, function_arg)
  else:
    raise utils.BayesDBError()

  return description




#
#   Copyright (c) 2010-2014, MIT Probabilistic Computing Project
#
#   Lead Developers: Jay Baxter and Dan Lovell
#   Authors: Jay Baxter, Dan Lovell, Baxter Eaves, Vikash Mansinghka
#   Research Leads: Vikash Mansinghka, Patrick Shafto
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#

import re
import utils
import numpy
import os
import pylab
import matplotlib.cm
import inspect
import operator
import ast
import math
from scipy.stats import pearsonr, chi2_contingency, f_oneway

import utils
import select_utils
import data_utils as du

###
# Three types of function signatures, for each purpose.
#
# SELECT/ORDER BY/WHERE:
# f(args, row_id, data_values, M_c, X_L_list, X_D_list, T, engine)
#
# ESTIMATE COLUMNS
#
#
# First argument of each of these functions is the function-specific argument list,
# which is parsed from parse_<function_name>(), also in this file.
#
# TODO: data_values unused
##

###################################################################
# NORMAL FUNCTIONS (have a separate output value for each row: can ORDER BY, SELECT, etc.)
###################################################################


def _column(column_args, row_id, data_values, M_c, X_L_list, X_D_list, T, engine, numsamples):
    col_idx = column_args[0]
    confidence = column_args[1]
    if confidence is None or not numpy.isnan(T[row_id][col_idx]):
        return du.convert_code_to_value(M_c, col_idx, T[row_id][col_idx])
    else:
        ## Do impute.
        Y = [(row_id, cidx, T[row_id][cidx]) for cidx in M_c['name_to_idx'].values() \
                   if not numpy.isnan(T[row_id][cidx])]
        code = utils.infer(M_c, X_L_list, X_D_list, Y, row_id, col_idx, numsamples,
                           confidence, engine)
        if code is not None:
            # Inferred successfully! Fill in the new value.
            value = du.convert_code_to_value(M_c, col_idx, code)
            return value
        else:
            return du.convert_code_to_value(M_c, col_idx, T[row_id][col_idx])

def _column_ignore(col_idx, row_id, data_values, M_c_full, T_full, engine):
    """
    This function handles selecting data from ignore columns. It's split into a different
    function because it needs to be passed M_c_full and T_full instead of M_c and T, as in _column.
    Since selecting ignore columns is probably a rare event, we can avoid passing M_c_full and T_full
    to _column as "just in case" arguments.
    """
    return du.convert_code_to_value(M_c_full, col_idx, T_full[row_id][col_idx])    

def _row_id(args, row_id, data_values, M_c, X_L_list, X_D_list, T, engine, numsamples):
    return row_id

def _similarity(similarity_args, row_id, data_values, M_c, X_L_list, X_D_list, T, engine, numsamples):
    target_row_id, target_columns = similarity_args
    return engine.call_backend('similarity', dict(M_c=M_c, X_L_list=X_L_list, X_D_list=X_D_list, given_row_id=row_id, target_row_id=target_row_id, target_columns=target_columns))

def _row_typicality(row_typicality_args, row_id, data_values, M_c, X_L_list, X_D_list, T, engine, numsamples):
    return engine.call_backend('row_structural_typicality', dict(X_L_list=X_L_list, X_D_list=X_D_list, row_id=row_id))

def _predictive_probability(predictive_probability_args, row_id, data_values, M_c, X_L_list, X_D_list, T, engine, numsamples):
    c_idx = predictive_probability_args
    assert type(c_idx) == int    
    Q = [(row_id, c_idx, T[row_id][c_idx])]
    Y = []
    p = math.exp(engine.call_backend('simple_predictive_probability_multistate', dict(M_c=M_c, X_L_list=X_L_list, X_D_list=X_D_list, Y=Y, Q=Q)))
    return p

#####################################################################
# AGGREGATE FUNCTIONS (have only one output value)
#####################################################################
    
def _col_typicality(col_typicality_args, row_id, data_values, M_c, X_L_list, X_D_list, T, engine, numsamples):
    c_idx = col_typicality_args
    assert type(c_idx) == int
    return engine.call_backend('column_structural_typicality', dict(X_L_list=X_L_list, col_id=c_idx))

def _probability(probability_args, row_id, data_values, M_c, X_L_list, X_D_list, T, engine, numsamples):
    c_idx, value = probability_args
    assert type(c_idx) == int
    try:
        observed = du.convert_value_to_code(M_c, c_idx, value)
    except KeyError:
        # value doesn't exist
        return 0
    row_id = len(X_D_list[0][0]) + 1 ## row is set to 1 + max row, instead of this row.
    Q = [(row_id, c_idx, observed)]
    Y = []
    p = math.exp(engine.call_backend('simple_predictive_probability_multistate', dict(M_c=M_c, X_L_list=X_L_list, X_D_list=X_D_list, Y=Y, Q=Q)))
    return p
    

#########################################################################
## TWO COLUMN AGGREGATE FUNCTIONS (have only one output value, and take two columns as input)
#########################################################################

def _dependence_probability(dependence_probability_args, row_id, data_values, M_c, X_L_list, X_D_list, T, engine, numsamples):
    """
    TODO: THIS NEEDS TO BE A FUNCTION ON CROSSCAT ENGINE! MOVE IT THERE!
    """
    col1, col2 = dependence_probability_args
    prob_dep = 0
    for X_L, X_D in zip(X_L_list, X_D_list):
        assignments = X_L['column_partition']['assignments']
        ## Columns dependent if in same view, and the view has greater than 1 category
        ## Future work can investigate whether more advanced probability of dependence measures
        ## that attempt to take into account the number of outliers do better.
        if (assignments[col1] == assignments[col2]):
            if len(numpy.unique(X_D[assignments[col1]])) > 1 or col1 == col2:
                prob_dep += 1
    prob_dep /= float(len(X_L_list))
    return prob_dep

def _old_dependence_probability(dependence_probability_args, row_id, data_values, M_c, X_L_list, X_D_list, T, engine, numsamples):
    col1, col2 = dependence_probability_args
    prob_dep = 0
    for X_L, X_D in zip(X_L_list, X_D_list):
        assignments = X_L['column_partition']['assignments']
        ## Columns dependent if in same view, and the view has greater than 1 category
        ## Future work can investigate whether more advanced probability of dependence measures
        ## that attempt to take into account the number of outliers do better.
        if (assignments[col1] == assignments[col2]):
            prob_dep += 1
    prob_dep /= float(len(X_L_list))
    return prob_dep

    
def _mutual_information(mutual_information_args, row_id, data_values, M_c, X_L_list, X_D_list, T, engine, numsamples):
    col1, col2 = mutual_information_args
    Q = [(col1, col2)]
    ## Returns list of lists.
    ## First list: same length as Q, so we just take first.
    ## Second list: MI, linfoot. we take MI.
    if numsamples is None:
        numsamples = 100
    # backend n_samples argument specifies samples per model
    n_samples = int(math.ceil(float(numsamples)/len(X_L_list)))
    results_by_model = engine.call_backend('mutual_information', dict(M_c=M_c, X_L_list=X_L_list, X_D_list=X_D_list, Q=Q, n_samples=n_samples))[0][0]
    
    ## Report the average mutual information over each model.
    mi = float(sum(results_by_model)) / len(results_by_model)
    return mi
    
def _correlation(correlation_args, row_id, data_values, M_c, X_L_list, X_D_list, T, engine, numsamples):
    col1, col2 = correlation_args

    # Create map of modeltype to column type. Treate cyclic as numerical for the purpose of calculating correlation.
    cctype_map = dict(normal_inverse_gamma = 'numerical', symmetric_dirichlet_discrete = 'categorical', vonmises = 'numerical')
    cctype1 = cctype_map[M_c['column_metadata'][col1]['modeltype']]
    cctype2 = cctype_map[M_c['column_metadata'][col2]['modeltype']]

    correlation = numpy.nan
    t_array = numpy.array(T, dtype=float)
    nan_index = numpy.logical_or(numpy.isnan(t_array[:,col1]), numpy.isnan(t_array[:,col2]))
    t_array = t_array[numpy.logical_not(nan_index),:]
    n = t_array.shape[0]

    if cctype1 == 'numerical' and cctype2 == 'numerical':
        # Two numerical columns: Pearson R squared
        correlation, p_value = pearsonr(t_array[:,col1], t_array[:,col2])
        correlation = correlation ** 2
    elif cctype1 == 'categorical' and cctype2 == 'categorical':
        # Two categorical columns: Cramer's phi
        data_i = numpy.array(t_array[:, col1], dtype='int32')
        data_j = numpy.array(t_array[:, col2], dtype='int32')
        unique_i = numpy.unique(data_i)
        unique_j = numpy.unique(data_j)
        min_levels = min(len(unique_i), len(unique_j))

        if min_levels >= 2:
            # Create contingency table - built-in way to do this?
            contingency_table = numpy.zeros((len(unique_i), len(unique_j)), dtype='int')
            for i in unique_i:
                for j in unique_j:
                    contingency_table[i][j] = numpy.logical_and(data_i == i, data_j == j).sum()

            chisq, p, dof, expected = chi2_contingency(contingency_table, correction=False)
            correlation = (chisq / (n * (min_levels - 1))) ** 0.5
    else:
        # One numerical, one categorical column: ANOVA R-squared
        if cctype1 == 'categorical':
            data_group = t_array[:, col1]
            data_y = t_array[:, col2]
        else:
            data_group = t_array[:, col2]
            data_y = t_array[:, col1]
        group_values = numpy.unique(data_group)
        n_groups = float(len(group_values))

        if n > n_groups:
            # Use scipy.stats.f_oneway to calculate F-statistic and p-value.
            F, p = f_oneway(*[data_y[data_group == j] for j in group_values])
            # Convert F-stat and number of groups into R-squared.
            correlation = 1 - (1 + F * ((n_groups - 1) / (n - n_groups))) ** -1

    return correlation
